@@ -48,7 +48,8 @@ class DynamoDBGlobalConfiguration:
 
     _instance = None
 
-    ddb: object
+    _ddb: object
+    _tables: dict
     use_aws_cli_credentials: bool
     access_params: dict
     single_table_name_env_var: str
@@ -60,6 +61,7 @@ class DynamoDBGlobalConfiguration:
     def get_instance(
         cls,
         ddb: object = None,
+        tables: dict = None,
         use_aws_cli_credentials: bool = True,
         access_params: dict = None,
         single_table_name_env_var: str = "DYNAMODB_TABLE_NAME",
@@ -107,7 +109,9 @@ class DynamoDBGlobalConfiguration:
                 If true an instance exists, it will be overwritten. Defaults to False.
         """
 
+
         if cls._instance is None or overwrite_existing_instance:
+            tables = tables if tables is not None else {}
             if single_table_inverse_index_properties is None:
                 single_table_inverse_index_properties = {
                     ConstantString.INVERSE_INDEX_NAME.value: "gsi1",
@@ -132,6 +136,7 @@ class DynamoDBGlobalConfiguration:
                                                                                    " to use the aws cli credentials")
             cls._instance = cls(
                 ddb,
+                tables,
                 use_aws_cli_credentials,
                 access_params,
                 single_table_name_env_var,
@@ -146,12 +151,12 @@ class DynamoDBGlobalConfiguration:
         return cls._instance is not None
 
     def get_connection(self):
-        if self.ddb is None: # Connect to the database if ddb is None
+        if self._ddb is None: # Connect to the database if ddb is None
             try:
                 if self.use_aws_cli_credentials:
-                    self.ddb = boto3.resource("dynamodb")
+                    self._ddb = boto3.resource("dynamodb")
                 else:
-                    self.ddb = boto3.resource(
+                    self._ddb = boto3.resource(
                         "dynamodb",
                         endpoint_url=self.access_params["endpoint_url"],
                         region_name=self.access_params["region_name"],
@@ -161,7 +166,21 @@ class DynamoDBGlobalConfiguration:
             except Exception as ex:
                 print(ex)
                 raise ConnectionRefusedError("Not able to connect to DynamoDB")
-        return self.ddb
+        return self._ddb
+
+    def get_table(self, table_name):
+        if table_name not in self._tables:
+            try:
+                self._tables[table_name] = self.get_connection().Table(table_name)
+            except Exception as ex:
+                print(ex)
+                warning_str = (
+                        "Could not access table "
+                        + str(table_name)
+                        + " check if the table exists"
+                )
+                raise ResourceWarning(warning_str)
+        return self._tables[table_name]
 
 
 ###### ---------- Global parameters ---------- ######
@@ -552,8 +571,6 @@ def _pk2id(pk):
 
 
 class DynamoCrudRepository(CrudRepository):
-
-    ddb = None
     table = None
     table_name: str = None
     map_dict: map = None
@@ -573,6 +590,8 @@ class DynamoCrudRepository(CrudRepository):
             dynamo_table_name(str, optional):
                 The name of the real table on DynamoDB
         """
+        if entity_type is None:
+            raise ValueError("You need to provide an entity_type class")
         self.entity_type = entity_type
 
         self.map_dict = self.entity_type.dynamo_map()
@@ -581,7 +600,6 @@ class DynamoCrudRepository(CrudRepository):
         global_values = DynamoDBGlobalConfiguration.get_instance()
         # For a single table design, define the table name as a global parameter,
         # if the table name is not set in the global parameters, will use the entity class table name
-        self.ddb = global_values.get_connection()
         if dynamo_table_name is not None:
             self.table_name = dynamo_table_name
         elif hasattr(entity_type, "dynamo_table_name") and entity_type.dynamo_table_name() is not None:
@@ -591,19 +609,17 @@ class DynamoCrudRepository(CrudRepository):
             self.table_name = self.entity_type.dynamo_table_name()
         else:
             import os
-            self.table_name = os.environ[DynamoDBGlobalConfiguration.get_instance().single_table_name_env_var]
-        try:
-            self.table = self.ddb.Table(self.table_name)
-        except Exception as ex:
-            print(ex)
-            warning_str = (
-                "Could not access table "
-                + str(self.table_name)
-                + " check if the table exists"
-            )
-            raise ResourceWarning(warning_str)
+            if global_values.single_table_name_env_var in os.environ:
+                self.table_name = os.environ[global_values.single_table_name_env_var]
+            else:
+                raise ValueError("You need to provide the dynamo table name")
+
+    def get_table_if_none(self):
+        if self.table is None:
+            self.table = DynamoDBGlobalConfiguration.get_instance().get_table(self.table_name)
 
     def count(self):
+        self.get_table_if_none()
         n_items = 0
 
         index_table_name = self.table_name if hasattr(self.entity_type, "dynamo_pk") else self.entity_type.dynamo_logical_table_name()
@@ -664,6 +680,7 @@ class DynamoCrudRepository(CrudRepository):
          If a list if provided, assumes the pattern [pk, sk]
         :return: an o object of the mapped class or a list of objects of the mapped class
         """
+        self.get_table_if_none()
         if isinstance(keys, list):  # Produce dict from list
             keys = self.key_list2key_map(keys)
         try:
@@ -707,6 +724,9 @@ class DynamoCrudRepository(CrudRepository):
             return self.remove_by_id(getattr(to_delete, to_delete.dynamo_id()))
 
     def item_attr2_instance_attr(self, item_attr_val, instance_attr_name: str, instance_attr_type: type):
+        if hasattr(instance_attr_type,  "__supertype__"):
+            instance_attr_type = instance_attr_type.__supertype__
+
         if ConstantString.ITEM2OBJ_CONV.value in self.map_dict[instance_attr_name]:
             return self.map_dict[instance_attr_name][ConstantString.ITEM2OBJ_CONV.value](item_attr_val)
         # convert to string then to the actual type to make sure the conversion will work
@@ -738,53 +758,16 @@ class DynamoCrudRepository(CrudRepository):
             return instance_attr_type(item_attr_val)
 
     def item2instance(self, item):
-        #entity_instance = self.entity_type()
         instance_attributes = {}
         for fl in fields(self.entity_type):
             item_attr_name = self.map_dict[fl.name][ConstantString.ITEM_ATTR_NAME.value]
             if item_attr_name in item:
                 instance_attributes[fl.name] = self.item_attr2_instance_attr(item[item_attr_name], fl.name, fl.type)
-                # if ConstantString.ITEM2OBJ_CONV.value in self.map_dict[fl.name]:
-                #     setattr(
-                #         entity_instance,
-                #         fl.name,
-                #         self.map_dict[fl.name][ConstantString.ITEM2OBJ_CONV.value](item[item_attr_name]),
-                #     )
-                # # convert to string then to the actual type to make sure the conversion will work
-                # elif issubclass(fl.type, (int, float, Decimal)):
-                #     setattr(entity_instance, fl.name, fl.type(str(item[item_attr_name])))
-                #
-                # elif issubclass(fl.type, (bytes, bool)):  # Perform a direct conversion
-                #     setattr(entity_instance, fl.name, fl.type(item[item_attr_name]))
-                #
-                # elif issubclass(fl.type, (dict, list)):  # json
-                #     setattr(entity_instance, fl.name, json.loads(item[item_attr_name]))
-                #
-                # elif issubclass(fl.type, Enum):  # Enum
-                #     setattr(entity_instance, fl.name, fl.type[item[item_attr_name]])
-                #
-                # elif issubclass(fl.type, str):
-                #     setattr(entity_instance, fl.name, str(item[item_attr_name]))
-                #
-                # # Use the iso format for storing datetime as strings
-                # elif issubclass(
-                #     fl.type, (datetime.date, datetime.time, datetime.datetime)
-                # ):
-                #     setattr(
-                #         entity_instance, fl.name, fl.type.fromisoformat(item[item_attr_name])
-                #     )
-                #
-                # elif issubclass(fl.type, object):  # objects in general are pickled
-                #     setattr(
-                #         entity_instance, fl.name, pickle.loads(bytes(item[item_attr_name]))
-                #     )
-                #
-                # else:  # No special case, use a simple cast, probably will never be reached
-                #     setattr(entity_instance, fl.name, fl.type(item[item_attr_name]))
         entity_instance = self.entity_type(**instance_attributes)
         return entity_instance
 
     def find_by_keys(self, keys: Union[dict, list]):
+        self.get_table_if_none()
         """
         Finds objects stored in the database using the given keys
         :param keys: dict, list a set of keys to search for the object.
@@ -856,6 +839,7 @@ class DynamoCrudRepository(CrudRepository):
             return None
 
     def find_by_id(self, unique_id):
+        self.get_table_if_none()
         return self.find_by_keys(self._id2key_pair(unique_id))
 
     def keys2KeyConditionExpression(self, keys: dict):
@@ -870,6 +854,7 @@ class DynamoCrudRepository(CrudRepository):
         return buffer, exp_att_val
 
     def find_all(self):
+        self.get_table_if_none()
         entity_list = []
 
         try:
@@ -985,58 +970,14 @@ class DynamoCrudRepository(CrudRepository):
         # Get every attribute of obj, ignoring private members and methods
         for instance_attr in inspect.getmembers(obj):
             if (
-                    (not instance_attr[0].startswith("_")) # ignore private attributes
-                    and (not inspect.ismethod(instance_attr[1])) # ignore methods
-                    and (not instance_attr[0] in obj.dynamo_ignore()) # ignore the attributes on the list to ignore
+                    (not instance_attr[0].startswith("_"))  # ignore private attributes
+                    and (not inspect.ismethod(instance_attr[1]))  # ignore methods
+                    and (not instance_attr[0] in obj.dynamo_ignore())  # ignore the attributes on the list to ignore
             ):
                 instance_attr_name = instance_attr[0]
                 instance_attr_val = instance_attr[1]
                 item_attr_val, item_attr_name = self.instance_attr2_item_attr(instance_attr_val, instance_attr_name)
                 item_params[item_attr_name] = item_attr_val
-                # item_attr_name = self.map_dict[instance_attr_name][ConstantString.ITEM_ATTR_NAME.value]
-                # if ConstantString.OBJ2ITEM_CONV.value in self.map_dict[instance_attr_name]:
-                #     item_params[item_attr_name] = self.map_dict[instance_attr_name][ConstantString.OBJ2ITEM_CONV.value](instance_attr_val)
-                # # switch self.map_dict[<attribute_name>]
-                # elif self.map_dict[instance_attr_name][ConstantString.ITEM_ATTR_TYPE.value] == "N":  # case 'N' (number)
-                #     item_params[item_attr_name] = Decimal(
-                #         str(instance_attr_val)
-                #     )  # str cast to support numpy, pandas, etc
-                #
-                # elif self.map_dict[instance_attr_name][ConstantString.ITEM_ATTR_TYPE.value] == "B":  # case 'B' (bytes)
-                #     if isinstance(instance_attr_val, bytes):
-                #         item_params[item_attr_name] = bytes(instance_attr_val)
-                #     elif isinstance(
-                #             instance_attr_val, object
-                #     ):  # objects in general are pickled
-                #         item_params[item_attr_name] = pickle.dumps(instance_attr_val)
-                #     else:
-                #         raise TypeError(
-                #             "Only bytes and objects should be stored as bytes"
-                #         )
-                # elif self.map_dict[instance_attr_name][ConstantString.ITEM_ATTR_TYPE.value] == "BOOL":  # case 'BOOL' (boolean)
-                #     item_params[item_attr_name] = 1 if instance_attr_val else 0
-                # else:  # default (string)
-                #     # Consider special cases and use specific string formats
-                #     # datetime
-                #     if isinstance(
-                #             instance_attr_val, (datetime.date, datetime.time, datetime.datetime)
-                #     ):
-                #         item_params[item_attr_name] = instance_attr_val.isoformat()
-                #
-                #     # enum
-                #     elif isinstance(instance_attr_val, Enum):
-                #         item_params[item_attr_name] = instance_attr_val.name
-                #
-                #     # maps and lists (converted to json)
-                #     elif isinstance(instance_attr_val, (dict, list)):
-                #         item_params[item_attr_name] = json.dumps(instance_attr_val)
-                #
-                #     # strings
-                #     elif isinstance(instance_attr_val, str):
-                #         item_params[item_attr_name] = str(instance_attr_val)
-                #     # No special case, use a simple str cast
-                #     else:
-                #         item_params[item_attr_name] = str(instance_attr_val)
         return item_params
 
     def instance2item_params(self, obj: T):
@@ -1046,6 +987,7 @@ class DynamoCrudRepository(CrudRepository):
         return item_params
 
     def save(self, obj: T):
+        self.get_table_if_none()
         item_params = self.instance2item_params(obj)
         try:
             self.table.put_item(Item=item_params)
@@ -1061,6 +1003,7 @@ class DynamoCrudRepository(CrudRepository):
             return False
 
     def remove_entity_list(self, entity_list: Iterable):
+        self.get_table_if_none()
         try:
             with self.table.batch_writer() as batch:
                 if hasattr(self.entity_type, "dynamo_pk"):
@@ -1093,6 +1036,7 @@ class DynamoCrudRepository(CrudRepository):
         return self.remove_entity_list(entity_list)
 
     def remove_all_by_keys(self, keys_list: Union[Iterable[dict], Iterable[list]]):
+        self.get_table_if_none()
         try:
             with self.table.batch_writer() as batch:
                 for keys in keys_list:
@@ -1141,6 +1085,7 @@ class DynamoCrudRepository(CrudRepository):
         return self.exists_by_keys(self._id2key_pair(unique_id))
 
     def save_all(self, entities: Iterable[T]):
+        self.get_table_if_none()
         try:
             with self.table.batch_writer() as batch:
                 for obj in entities:
@@ -1155,6 +1100,8 @@ class DynamoCrudRepository(CrudRepository):
     @classmethod
     def dynamo_type_from_type(cls, python_type: type):
         # if using a specific library like numpy or pandas, the user should specify the "N" type himself
+        if hasattr(python_type, "__supertype__"):
+            python_type = python_type.__supertype__
         if issubclass(python_type, (int, float, Decimal)):
             dynamo_type = "N"
         elif issubclass(
@@ -1218,7 +1165,11 @@ class SingleTableDynamoCrudRepository(DynamoCrudRepository):
                 dynamo_table_name = entity_type.dynamo_table_name()
             else:
                 import os
-                dynamo_table_name = os.environ[DynamoDBGlobalConfiguration.get_instance().single_table_name_env_var]
+                global_values = DynamoDBGlobalConfiguration.get_instance()
+                if global_values.single_table_name_env_var in os.environ:
+                    dynamo_table_name = os.environ[global_values.single_table_name_env_var]
+                else:
+                    raise ValueError("You need to provide the dynamo table name")
         super().__init__(entity_type, dynamo_table_name)
 
     def instance2item_params_inject_keys(self, obj: T, item_params: dict = None):
@@ -1237,6 +1188,7 @@ class SingleTableDynamoCrudRepository(DynamoCrudRepository):
         return item_params
 
     def count(self):
+        self.get_table_if_none()
         try:
             inverse_index = DynamoDBGlobalConfiguration.get_instance().single_table_inverse_index_properties
             response = self.table.query(
@@ -1260,6 +1212,7 @@ class SingleTableDynamoCrudRepository(DynamoCrudRepository):
         return self.find_by_keys(self._id2key_pair(unique_id))
 
     def find_all(self):
+        self.get_table_if_none()
         entity_list = []
 
         try:
@@ -1283,6 +1236,7 @@ class SingleTableDynamoCrudRepository(DynamoCrudRepository):
         return entity_list
 
     def remove_entity_list(self, entity_list: Iterable):
+        self.get_table_if_none()
         try:
             with self.table.batch_writer() as batch:
                 for entity in entity_list:
@@ -1331,7 +1285,11 @@ class MultiTableDynamoCrudRepository(DynamoCrudRepository):
                 dynamo_table_name = entity_type.table_name()
             else:
                 import os
-                dynamo_table_name = os.environ[DynamoDBGlobalConfiguration.get_instance().single_table_name_env_var]
+                global_values = DynamoDBGlobalConfiguration.get_instance()
+                if global_values.single_table_name_env_var in os.environ:
+                    dynamo_table_name = os.environ[global_values.single_table_name_env_var]
+                else:
+                    raise ValueError("You need to provide the dynamo table name")
         if not hasattr(entity_type, "dynamo_pk"):  # Define dynamo_pk using the unique_id
             cls_map = entity_type.dynamo_map()
             if entity_type.dynamo_id() in cls_map and ConstantString.ITEM_ATTR_NAME.value in cls_map[entity_type.dynamo_id()]:
@@ -1370,6 +1328,7 @@ class MultiTableDynamoCrudRepository(DynamoCrudRepository):
             item_params[obj.dynamo_pk_name_on_table()] = str(getattr(obj, obj.dynamo_id(), ""))
 
     def count(self):
+        self.get_table_if_none()
         try:
             inverse_index = DynamoDBGlobalConfiguration.get_instance().single_table_inverse_index_properties
             response = self.table.query(
@@ -1414,6 +1373,7 @@ class MultiTableDynamoCrudRepository(DynamoCrudRepository):
         return self.find_by_keys(self._id2key_pair(unique_id))
 
     def find_all(self):
+        self.get_table_if_none()
         entity_list = []
 
         try:
@@ -1437,6 +1397,7 @@ class MultiTableDynamoCrudRepository(DynamoCrudRepository):
         return entity_list
 
     def remove_entity_list(self, entity_list: Iterable):
+        self.get_table_if_none()
         try:
             with self.table.batch_writer() as batch:
                 for entity in entity_list:
